@@ -253,6 +253,8 @@ namespace heif {
 
     int peek_bits(int n);
 
+    void skip_bytes(int nBytes);
+
     void skip_bits(int n);
 
     void skip_bits_fast(int n);
@@ -268,6 +270,11 @@ namespace heif {
       return data_length - bytes_remaining - nextbits_cnt / 8;
     }
 
+    int64_t get_bits_remaining() const
+    {
+      return bytes_remaining*8 + nextbits_cnt;
+    }
+
   private:
     const uint8_t* data;
     int data_length;
